#include "Args.h"

#include <ctype.h>
#include <strstream.h>

// note:
//
// might let a bare dash ("-") guide to the default arg.
//


const HelpArg Args::ourHelp;

// defines templated code for linking.
template class TArg<int>;
template class TArg<double>;
template class TArg<string>;

template <class T>
bool TArg<T>::parseArgs(int & argc, char **& argv, bool negative)
{
  if (argc == 0)
    return false;

  istrstream in(argv[0]);

  if (!(in >> val))
    return false;

  argc--;
  argv++;

  return true;
}

template <class T>
void TArg<T>::printUsage(ostream & out)
{
  out << "<" << myArgString << ">";
}

template <class T>
void TArg<T>::printExplanation(ostream & out)
{
  out << myExpl << " (default = \"" << myDefault << "\")";
}

Args::Args(bool standAlone, bool withHelp) : 
  myNumRequired(0), myDefaultArg(0), myDefaultRequired(false), 
  myStandAlone(standAlone)
{
  for (int i = 0; i < numArgs; i++)
    myArgs[i] = 0;
  
  if (withHelp)
    addArg('h', "help", (Arg*)&ourHelp);
}

bool Args::addArg(char flagLetter, Arg * arg, bool required)
{
  if (arg == 0 || isspace(flagLetter) || !isprint(flagLetter) ||
	 flagLetter == '-' || flagLetter == '+' || myArgs[flagLetter] != 0)
    return false;
  
  myArgsRequired[flagLetter] = required;
  myArgs[flagLetter] = arg;

  if (required)
    myNumRequired++;

  return true;
}

bool Args::addArg(char flagLetter, string flagString, Arg * arg,
			   bool required)
{
  if (flagString.length() == 0 || 
	 myFullFlags.find(flagString) != myFullFlags.end())
    return false;

  if (!addArg(flagLetter, arg, required))
    return false;

  myFullFlags[flagString] = flagLetter;
}

bool Args::addDefaultArg(Arg * arg, bool required)
{
  if (myDefaultArg != 0 || arg == 0)
    return false;

  myDefaultArg = arg;
  myDefaultRequired = required;

  return true;
}

Arg * Args::getArg(char flagLetter)
{
  return myArgs[flagLetter];
}

Arg * Args::getArg(string flagString)
{
  if (myFullFlags.find(flagString) == myFullFlags.end())
    return NULL;

  return getArg(myFullFlags[flagString]);
}

Arg * Args::getDefaultArg()
{
  return myDefaultArg;
}

void Args::printUsage(ostream & out)
{
  hash_map<string, char>::iterator mi;
  bool printed[numArgs] = { false };

  printHeader(out);

  // first print required arguments
  for (mi = myFullFlags.begin(); mi != myFullFlags.end(); mi++) {
    if (myArgs[(*mi).second] != 0 && myArgsRequired[(*mi).second]) {
	 out << endl;
	 printArg(out, (*mi).second, myArgs[(*mi).second], true, (*mi).first);
	 printed[(*mi).second] = true;
    }
  }

  for (int i = 0; i < numArgs; i++) {
    if (myArgs[i] != 0 && myArgsRequired[i] && !printed[i]) {
	 out << endl;
	 printArg(out, (char)i, myArgs[i], true);
	 printed[i] = true;
    }
  }

  // then print optional arguments
  for (mi = myFullFlags.begin(); mi != myFullFlags.end(); mi++) {
    if (myArgs[(*mi).second] != 0 && !printed[(*mi).second]) {
	 out << endl;
	 printArg(out, (*mi).second, myArgs[(*mi).second], false, (*mi).first);
	 printed[(*mi).second] = true;
    }
  }

  for (int i = 0; i < numArgs; i++) {
    if (myArgs[i] != 0 && !printed[i]) {
	 out << endl;
	 printArg(out, (char)i, myArgs[i], false);
	 printed[i] = true;
    }
  }

  // finally, print default argument
  if (myDefaultArg) {
    out << endl;
    printArg(out, 0, myDefaultArg, myDefaultRequired);
  }

  out << endl;
}

void Args::printHeader(ostream & out)
{
  if (myStandAlone) {
    out << "Usage (use '-' for false, '+' for true on switch flags):";
  }
}

void Args::printArg(ostream & out, char letter, Arg * arg, bool req,
				string str)
{
  out << "  ";

  if (!req)
    out << "[";

  if (str != "") {
    out << "(" << "--" << str << "|";
  }

  if (letter != 0)
    out << "-" << letter;

  if (str != "") {
    out << ")";
  }

  out << " ";
  arg->printUsage(out);

  if (!req)
    out << "] ";

  out << "\t-- "; 
  arg->printExplanation(out);
}

void Args::error(string errorMsg)
{
  if (!myStandAlone)
    return;

  cerr << "Error in arguments";
  if (errorMsg != "") {
    cerr << ": " << errorMsg;
  }
  cerr << endl;

  printUsage(cerr);
}

bool Args::parseArgs(int & argc, char **& argv, bool negative)
{
  int required = myNumRequired;
  bool neg;
  char flagLetter;
  char * current;
  bool usedFlags[numArgs] = { false };

  if (myStandAlone) {
    argc--;
    argv++;
  }

  while (argc && (argv[0][0] == '-' || argv[0][0] == '+')) {
    current = argv[0];

    neg = (argv[0][0] == '-');
    int arglen = strlen(argv[0]);

    if (arglen == 1) {
	 error(((string)"Invalid argument (no flag specified): ") + current);
	 return false;
    }

    // string style argument
    if (argv[0][0] == argv[0][1]) {
	 string flagString = argv[0] + 2;

	 if (arglen == 2) {
	   error(((string)"Invalid argument (no string specified): ") + current);
	   return false;
	 }

	 if (myFullFlags.find(flagString) == myFullFlags.end()) {
	   error(((string)"Unknown argument: ") + current);
	   return false;
	 }

	 flagLetter = myFullFlags[flagString];
    }
    // char style argument
    else {
	 if (arglen > 2) {
	   error(((string)"Flag \"") + current + "\" does not make sense." + 
		    " Perhaps you want \"" + current[0] + current + "\" or \"" +
		    current[0] + current[1] + "\"?");
	   return false;
	 }

	 flagLetter = argv[0][1];
    }

    if (myArgs[flagLetter] == 0) {
	 error(((string)"Flag \"") + current + "\" does not exist.");
	 return false;
    }

    if (usedFlags[flagLetter]) {
	 error(((string)"Flag \"") + flagLetter + "\" set more than once.");
	 return false;
    }
    usedFlags[flagLetter] = true;

    argc--;
    argv++;

    if (!myArgs[flagLetter]->parseArgs(argc, argv, neg)) {
	 error(((string)"Problem with arguments to flag \"") + current + "\"");
	 return false;
    }

    if (myArgsRequired[flagLetter])
	 required--;
  }

  if (required > 0) {
    bool first = true;
    string reqArgs = "";
    for (int i = 0; i < numArgs; i++) {
	 if (myArgs[i] && myArgsRequired[i] && !usedFlags[i]) {
	   if (first)
		first = false;
	   else
		reqArgs += " ";
	   reqArgs += ((string)"-") + (char)i;
	 }
    }
    error("Missing required arguments: " + reqArgs);
    return false;
  }

  if (argc > 0) {
    if (myDefaultArg == 0) {
	 error("Encountered unexpected trailing argument");
	 return false;
    }

    myDefaultArg->parseArgs(argc, argv, true);
  }
  else if (myDefaultRequired) {
    error("Required trailing argument missing");
    return false;
  }

  if (argc > 0 && myStandAlone) {
    error("Unhandled trailing arguments remain");
    return false;
  }

  return true;
}
