#ifndef _ARGS_H
#define _ARGS_H

#include <string>
#include <hash_map>

#ifndef HASH_STRING
#define HASH_STRING
struct hash<string>
{
  size_t operator()(const string & s) const {
    unsigned int i;
    size_t result = 0;
    for (i = 0; i < s.length(); i++) {
	 result <<= 1;
	 result += (size_t)s[i];
    }

    return result;
  }
};
#endif

const int numArgs = 256;

class Arg {
public:

  virtual bool parseArgs(int & argc, char **& argv, bool negative) = 0;
  virtual void printUsage(ostream & out) = 0;
  virtual void printExplanation(ostream & out) = 0;
  virtual ~Arg() {}

protected:
};

class HelpArg : public Arg {
public:
  bool parseArgs(int & argc, char **& argv, bool negative) { return false; }
  void printUsage(ostream & out) { }
  void printExplanation(ostream & out) { out << "Print usage statement."; }
};

class SwitchArg : public Arg {
public:
  SwitchArg(const string & expl, bool def) : 
    val(def), myExpl(expl), myDef(def) {};

  bool parseArgs(int & argc, char **& argv, bool negative) {
    val = !negative;

    return true;
  }

  void printUsage(ostream & out) { 
    // Nothing to do here
  }

  void printExplanation(ostream & out) {
    out << myExpl << " (default = " << (myDef ? "true" : "false") << ")";
  }

  bool val;

private:
  string myExpl;
  bool myDef;
};

template <class T>
class TArg : public Arg {
public:
  TArg(string arg, const T & def = T(), 
	  string expl = "Replace \"<...>\" with argument") : 
    val(def), myArgString(arg), myDefault(def), myExpl(expl) {}

  bool parseArgs(int & argc, char **& argv, bool negative);
  void printUsage(ostream & out);
  void printExplanation(ostream & out);

  T val;

private:
  string myArgString;
  T myDefault;
  string myExpl;
};

typedef TArg<string> StringArg;
typedef TArg<int> IntArg;
typedef TArg<double> DoubleArg;

class Args : public Arg {
public:
  Args(bool standAlone = true, bool withHelp = true);

  bool parseArgs(int & argc, char **& argv, bool negative = false);
  bool addArg(char flagLetter, Arg * arg, bool required = false);
  bool addArg(char flagLetter, string flagString, Arg * arg, 
		    bool required = false);
  bool addDefaultArg(Arg * arg, bool required = false);

  void printUsage(ostream & out);
  void printExplanation(ostream & out) {}

  Arg * getArg(char flagLetter);
  Arg * getArg(string flagString);
  Arg * getDefaultArg();

private:
  void printArg(ostream & out, char letter, Arg * arg, bool req, 
			 string str = "");
  void printHeader(ostream & out);
  void error(string errorMsg);

  Arg * myArgs[numArgs];
  bool myArgsRequired[numArgs];
  hash_map<string, char> myFullFlags;
  int myNumRequired;
  Arg * myDefaultArg;
  bool myDefaultRequired;
  bool myStandAlone;

  static const HelpArg ourHelp;
};


#endif // _ARGS_H
