//
// Take an unmapped solution and pretty print it
// Assumes a solution format as follows:
//
// UNSAT alone in file means unsatisfiable problem
// 
// Otherwise, only true vars matter. Looks for vars of the form:
// 
// {TIME_PREFIX}{INT}{CONNECTOR}{ACTION_LABEL}{action name}({CONNECTOR}{arg})*
//   ({CONNECTOR}{fluent name})*
//
// where {fluent name} has the ? and the action ID num but not the time step
// ID num.
//
// In the rvars, look for all non-action-ID fluents and floats. Collect stepped
// fluents together. Look for only the needed action params (the IDed fluents)
//

#include "shared.h"
#include <hash_map>
#include <string>
#include <iostream.h>
#include <stdlib.h>

typedef hash_map<string, double> sd_map;

struct proc_action_type {
  proc_action_type(string name_, int step_, vector<string> params_, 
			    vector<string> flparams_) :
    name(name_), step(step_), params(params_), flparams(flparams_) {}

  string name;
  int step;
  vector<string> params;
  vector<string> flparams;
};

int max_step = 0;
vector<proc_action_type> actions;
sd_map rmap;

void record_length(string token)
{
  int step;
  unsigned int pos, endpos;

  pos = token.find(TIME_PREFIX) + strlen(TIME_PREFIX);
  for (endpos = pos + 1; 
	  endpos < token.length() && token[endpos] >= '0' && token[endpos] <= '9';
	  endpos++) {
    ;
  }

  if (pos > token.length() || endpos > token.length())
    return;

  step = atoi(token.substr(pos, endpos).c_str());

  if (step > max_step)
    max_step = step;
}

bool match(string & source, string target)
{
  if (source.compare(target, 0, target.length()) != 0)
    return false;

  source = source.substr(target.length());

  return true;
}

// determines whether a given rvar is a float
bool is_float(string token)
{
  unsigned int pos = token.find(TIME_PREFIX);

  return (pos >= token.length());
}

// true only if the fluent is of the given step
bool is_fluent_step(string token, int step_target)
{
  unsigned int pos = token.find(TIME_PREFIX) + strlen(TIME_PREFIX);
  unsigned int endpos = token.find(VAR_INTERNAL_CONNECTOR);
  int step;

  if (pos > token.length() || endpos > token.length())
    return false;

  step = atoi(token.substr(pos, endpos).c_str());

  if (step != step_target)
    return false;

  token = token.substr(endpos + strlen(VAR_INTERNAL_CONNECTOR));
  assert(token.length() > 0);

  // otherwise, if it isn't numerically labeled _after_ the 
  // time label, it's a fluent (params are numerically labeled)
  return (token[0] < '0' || token[0] > '9');
}

bool is_action(string token)
{
  if (!match(token, TIME_PREFIX))
    return false;

  while (token.length() >= 1 && token[0] >= '0' && token[0] <= '9')
    token = token.substr(1);

  if (!match(token, CONNECTOR))
    return false;

  return match(token, ACTION_LABEL);
}

void process(string action)
{
  unsigned int pos;
  int time_step;
  string time_string;
  string param;
  string action_name;
  vector<string> params, fluent_params;
  vector< string > proc_flparams;
  vector<string>::iterator i;

  pos = action.find(CONNECTOR);
  time_string = action.substr(0, pos);
  match(action, TIME_PREFIX);
  time_step = atoi(action.substr(0, pos).c_str());
  pos = action.find(CONNECTOR);
  action = action.substr(pos + strlen(CONNECTOR));

  match(action, ACTION_LABEL);
  pos = action.find(CONNECTOR);
  if (pos > action.length())
    pos = action.length();

  action_name = action.substr(0, pos);
  action = action.substr(pos);

  while (match(action, CONNECTOR) && action.length() >= 0) {
    pos = action.find(CONNECTOR);
    if (pos > action.length()) {
	 pos = action.length();
    }

    param = action.substr(0, pos);
    action = action.substr(pos);

    if (param[0] == '?')
	 fluent_params.push_back(param);
    else
	 params.push_back(param);
  }

  for (i = fluent_params.begin(); i != fluent_params.end(); i++) {
    string full_rvar;
    full_rvar = time_string + (string)VAR_INTERNAL_CONNECTOR + (*i).substr(1);

    proc_flparams.push_back(full_rvar);
  }

  actions.push_back(proc_action_type(action_name, time_step, 
							  params, proc_flparams));
}

bool is_interesting(string rvar)
{
  string root;
  
  // all floats are interesting
  if (rvar.find(TIME_PREFIX) >= rvar.length())
    return true;

  root = rvar.substr(1 + strlen(TIME_PREFIX));

  while (root[0] >= '0' && root[0] <= '9')
    root = root.substr(1);

  assert(root[0] == VAR_INTERNAL_CONNECTOR[0] && 
	    strlen(VAR_INTERNAL_CONNECTOR) == 1);
  root = root.substr(1);

  // all fluents are interesting
  if (!(root[0] >= '0' && root[0] <= '9'))
    return true;

  // Now, only the parameters are left. Only parameters mentioned
  // in an interesting action are interesting.
  for (vector<proc_action_type>::iterator i = actions.begin();
	  i != actions.end(); i++) {
    for (vector<string>::iterator j = (*i).flparams.begin();
	    j != (*i).flparams.end(); j++) {
	 if ((*j) == rvar) {
	   return true;
	 }
    }
  }

  return false;
}

void rprocess(string rvar, double value)
{
  rmap[rvar] = value;
}

void prepare_solution(hash_map<string, bool> & soln, 
				  hash_map<string, double> & vsoln)
{
  hash_map<string, bool>::iterator mbi;
  hash_map<string, double>::iterator mdi;
  string token;
  bool val;

  for (mbi = soln.begin(); mbi != soln.end(); mbi++) {
    token = (*mbi).first;
    val = (*mbi).second;

    record_length(token);

    if (!val)
	 continue;

    if (is_action(token)) {
	 process(token);
    }
  }
  
  for (mdi = vsoln.begin(); mdi != vsoln.end(); mdi++) {
    if (is_interesting((*mdi).first)) {
	 rprocess((*mdi).first, (*mdi).second);
    }
  }
}

void output_float(string rvar, double value)
{
  cout << '\t' << rvar << " = " << value << endl;
}

void output_fluent(string rvar, double value)
{
  string root;

  root = "?" + rvar.substr(rvar.find(VAR_INTERNAL_CONNECTOR) + 
					  strlen(VAR_INTERNAL_CONNECTOR));

  output_float(root, value);
}

void output_floats()
{
  sd_map::iterator i;

  for (i = rmap.begin(); i != rmap.end(); i++) {
    if (is_float((*i).first)) {
	 output_float((*i).first, (*i).second);
    }
  }
}

void output_fluents(int step)
{
  sd_map::iterator i;

  for (i = rmap.begin(); i != rmap.end(); i++) {
    if (is_fluent_step((*i).first, step)) {
	 output_fluent((*i).first, (*i).second);
    }
  }
}

void output_actions(int step)
{
  unsigned int i;
  unsigned int j;

  for (i = 0; i < actions.size(); i++) {
    if (actions[i].step != step)
	 continue;

    cout << "\t" << actions[i].name;

    for (j = 0; j < actions[i].params.size(); j++) {
	 cout << " " << actions[i].params[j];
    }

    for (j = 0; j < actions[i].flparams.size(); j++) {
	 cout << " " << rmap[actions[i].flparams[j]];
    }

    cout << endl;
  }
}

void pretty_print_solution()
{
  int step;

  cout << "Float variables: " << endl;
  output_floats();

  for (step = 0; step < max_step; step++) {
    cout << endl << "Fluents at step " << step + 1 << ":" << endl;
    output_fluents(step);
    cout << endl << "Actions at step " << step + 1 << ":" << endl;
    output_actions(step);
  }

  cout << endl << "Fluents at step " << max_step + 1 << ":" << endl;
  output_fluents(max_step);

  cout << endl;
}

void pretty_print(bool solved, 
			   hash_map<string, bool> & soln, 
			   hash_map<string, double> & vsoln)
{
  if (solved) {
    prepare_solution(soln, vsoln);
    pretty_print_solution();
  }
  else {
    cout << "No plan found." << endl;
  }
}

