/////////////////
// OS Includes
#include <fstream.h>

//////////////
// Includes
#include "Clause.h"
#include "Random.h"
#include "SATInstance.h"
#include "SATSolver.h"
#include "creader.h"
#include "shared.h"
#include <strstream.h>
#include "Cl.h"

/////////////
// Defines

/////////////////////////////
// Static data initialization 
SATInstance SATInstance::xSATInstance;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Public Methods

void SATInstance::vOutput()
{
  for (int i=0; i<_iClauseCount; i++) {
    Clause* pPrintMe = _aClause[i];
    for (int j=0; j < pPrintMe->iVariableCount(); j++) {
      if (pPrintMe->bIsNegated(j))
	cout << "-";
      cout << (int) (pPrintMe->eConstrainedVariable(j)+1) << " "; 
    }
    cout << '\n';
  }
  cout << "%\n";
}

void SATInstance::vOutputDimacs()
{
  cout << "p cnf " << iVariableCount << ' ' << _iClauseCount << '\n';
  for (int i=0; i<_iClauseCount; i++) {
    Clause* pPrintMe = _aClause[i];
    for (int j=0; j < pPrintMe->iVariableCount(); j++) {
      if (pPrintMe->bIsNegated(j))
	cout << "-";
      cout << (int) (pPrintMe->eConstrainedVariable(j)+1) << " "; 
    }
    cout << " 0\n";
  }
  cout << flush;
}

boolean SATInstance::bReadDimacs(char* aFileName_)
{
  ifstream xInputFile(aFileName_);
  return bReadDimacs(xInputFile);
}

boolean SATInstance::bReadDimacs(istream& xInputFile)
{
  const int MAX_BUF_LEN = 5000;
  char buffer[MAX_BUF_LEN];
  VariableID eMaxVar = 0;
  int iClauseCount;
  char cCheck;
  boolean bConstraintSet = false;

  VariableSet xVariables;
  VariableSet xNegatedVariables;
  while (1) {
    xInputFile >> cCheck;
    if (cCheck == 'c') {
      xInputFile.getline(buffer, MAX_BUF_LEN);
      continue;
    }
    else if (cCheck == 'p') {
      xInputFile >> buffer;
      xInputFile >> eMaxVar;
      xInputFile >> iClauseCount;
      if (eMaxVar > MAX_VARIABLES) {
	   cout << "Error: Maximum variables allowed is " 
		   << MAX_VARIABLES << ".\n";
	   return 0;
      }

	 if (strcmp(buffer, "lcnf") == 0) {
	   xInputFile >> iRVariableCount;
	   bConstraintSet = true;
	 }
	 else {
	   iRVariableCount = 0;
	 }
      break;
    }
    else {
      cout << "Error: File not in DIMACS format.\n";
      return 0;
    }
  }
  iVariableCount = eMaxVar;
  int iWorkCount = 0;
  while (iWorkCount < iClauseCount) {
    if (xInputFile >> cCheck)
	 if (cCheck == 'c') {
	   xInputFile.getline(buffer, MAX_BUF_LEN);
	   continue;
	 }
	 else xInputFile.putback(cCheck);
    
    xVariables.vClear();
    xNegatedVariables.vClear();
    do {
      VariableID eVar = 0;
      xInputFile >> eVar;
      if (eVar == 0)
	   break;
      if(eVar > eMaxVar){
	   cout << "Error: some variable is numbered larger than the maximum.\n";
	   return 0;
      }
      if (eVar < 0) {
	   if (xVariables.bAddVariable(0-(eVar+1)) == 0) {
		cout << "Error: Some clause contains duplicate variables.\n";
		cout << eVar << endl;
		return 0;
	   }
	   xNegatedVariables.vAddVariable(0-(eVar+1));
      }
      else {
	   if (xVariables.bAddVariable(eVar-1) == 0) {
		cout << "Error: Some clause contains duplicate variables.\n";
		return 0;
	   }
      }
    } while (1);
    if (xVariables.iVariableCount() == 0) {
      cout << "Error: encountered a 0 length clause \n"; 
      return 0;
    }

    Clause* pNewConstraint = new Clause(xVariables, xNegatedVariables, 1);
    SATInstance::xSATInstance.vAddClause(pNewConstraint);
    iWorkCount++;
  }

  aConstraints = new ClConstraint*[iVariableCount];

  for (int i = 0; i < iVariableCount; i++) {
    aConstraints[i] = 0;
  }

  if (bConstraintSet) {
    aVariables = new ClVariable[iRVariableCount];

    // Constraints come one to a line in the form:
	 // ID Constraint
	    // Whitespace separates ID and Constraint.
	    // Terminated by an ID of 0.
	    // No variable should be constrained more than once.
	    VariableID eCurrent;
    while (xInputFile >> eCurrent && eCurrent != 0) {
	 if (eCurrent > eMaxVar) {
	   cout << "Error: some variable in the constraints section, " 
		   << eCurrent << ", is numbered larger than the maximum, " <<
		eMaxVar << "." << endl;
	   return 0;
	 }
	 
	 if (eCurrent < 0) {
	   cout << "Error: some variable in the constraints section, " 
		   << eCurrent << ", is negative (not currently supported)."
		   << endl;
	   return 0;
	 }
	 if (aConstraints[eCurrent-1] != 0) {
	   cout << "Warning: some variable in the constraints section, " 
		   << eCurrent << ", occurs more than once." << endl;
	   cout << "Ignoring all but first occurrence!" << endl;
	   continue;
	 }

	 char cBracket;

	 xInputFile >> cBracket;
	 
	 if (cBracket != '[') {
	   cerr << "Error: syntax error (missing '[')." << endl;
	   return 0;
	 }

	 xInputFile.getline(buffer, ']');

	 if (strlen(buffer) == 0 || buffer[strlen(buffer) - 1] != ']') {
	   cerr << "Error: syntax error in constraint specification." << endl;
	   return 0;
	 }

	 // The trailing ']' will still be on there. Get rid of it:
	   buffer[strlen(buffer) - 1] = 0;
	 istrstream in(buffer);

	 aConstraints[eCurrent-1] = parseConstraint(in, aVariables);

	 if (aConstraints[eCurrent-1] == 0) {
	   cout << "Warning: unable to parse constraint \"" << buffer 
		   << "\" for variable " << eCurrent << "." << endl;
	   cout << "Skipping." << endl;
	   continue;
	 }

	 aConstraints[eCurrent-1]->setPv((void*)(eCurrent - 1));
    }
  }

  return 1;
}

boolean SATInstance::bInit(const vector<vector<VAR_OCC> > &clauses, int numLit,
					  const vector<pair<int,string> > & constraints,
					  int numRVar)
{
  VariableID eMaxVar = -1;

  VariableSet xVariables;
  VariableSet xNegatedVariables;

  for (vector< vector<VAR_OCC> >::const_iterator ci = clauses.begin(); 
	  ci != clauses.end(); 
	  ci++) {
    xVariables.vClear();
    xNegatedVariables.vClear();
    
    for (vector<VAR_OCC>::const_iterator j = (*ci).begin(); 
	    j != (*ci).end(); j++) {
      VariableID eVar = (*j).val;

      if(eVar > eMaxVar){
	   eMaxVar = eVar;
      }

	 if (xVariables.bAddVariable(eVar) == 0) {
	   cerr << "Error: Some clause contains duplicate variables.\n";
	   return 0;
	 }

	 if (!(*j).pos) {
	   xNegatedVariables.vAddVariable(eVar);
      }
    }

    if (xVariables.iVariableCount() == 0) {
      cerr << "Error: encountered a 0 length clause \n";
      return 0;
    }

    Clause* pNewConstraint = new Clause(xVariables, xNegatedVariables, 1);
    SATInstance::xSATInstance.vAddClause(pNewConstraint);
  }

  assert(numLit == eMaxVar + 1);
  iVariableCount = eMaxVar + 1;
  aConstraints = new ClConstraint*[iVariableCount];

  for (int i = 0; i < iVariableCount; i++) {
    aConstraints[i] = 0;
  }

  iRVariableCount = numRVar;
  aVariables = new ClVariable[iRVariableCount];

  for (vector<pair<int,string> >::const_iterator cons = constraints.begin();
	  cons != constraints.end();
	  cons++) {
    VariableID var = (*cons).first;
    string cons_str = (*cons).second;
    
    if (var > eMaxVar) {
	 cerr << "Error: some variable in the constraints section, ";
	 cerr << var << ", is numbered larger than the maximum, "
		 << eMaxVar << ".\n";
	 return 0;
    }
  
    if (aConstraints[var] != 0) {
	 cerr << "Warning: some variable in the constraints section, ";
	 cerr << var << ", occurs more than once.\n";
	 cerr << "Ignoring all but first occurrence!\n";
	 continue;
    }

    istrstream in(cons_str.c_str());

    aConstraints[var] = parseConstraint(in, aVariables);
    
    if (aConstraints[var] == 0) {
	 cerr << "Warning: unable to parse constraint \"" 
		 << cons_str.c_str() << "\" for variable "  
	      << var << ".\n";
	 cerr << "Skipping.\n";
	 continue;
    }

    aConstraints[var]->setPv((void*)(var));
  }
  return 1;  
}

boolean SATInstance::bInit(const vector< vector<VAR_OCC> >&clauses, int maxLit)
{
  vector<pair<int, string> > constraints;
  return bInit(clauses, maxLit, constraints, 0);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Protected Methods

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Private Methods
