/////////////////
// OS Includes
#include <stdlib.h>
#include <fstream.h>


//////////////
// Includes
#include "ipp.h"
#include "Clause.h"
#include "Random.h"
#include "SATInstance.h"
#include "SATSolver.h"
#include "Cl.h"

/////////////////////////////
// Static Data Initialization

ClSimplexSolver VariableStruct::xSolver;
SATSolver *     VariableStruct::pSATSolver;
bool            VariableStruct::_bAutoFlush = true;
VariableSet     VariableStruct::xExplanation;
list<VariableStruct*> VariableStruct::_xCache;
int                   VariableStruct::_iCacheSize = 0;

double SATSolver::scoreCutoffFactor = 1;
double SATSolver::bestScoreFudgeFactor = 1;
double SATSolver::constraintPreferenceFactor = 1;
double SATSolver::constraintSizeFactor = 0;
bool   SATSolver::fastUnitPropConstrained = true;
bool   SATSolver::bExplanation = true;
bool   SATSolver::bNaiveSolver = true;
bool   SATSolver::bConstrainedVarsNegFirst = false;
bool   SATSolver::bUsePureLitElim = false;
int    SATSolver::iConsMaxLearnSize = -1;
int    SATSolver::iLogicMaxLearnSize = -1;
bool   SATSolver::bLearn = false;
bool   SATSolver::bRelevance = false;
bool   SATSolver::bTimeLimit = false;
int    SATSolver::iMaxTime = 0;
int    SATSolver::iCachePolicy = NO_CACHE;
int    SATSolver::iCacheDepth = 0;

/////////////
// Defines
#define MAX_TO_TRY                  10

//#define DEBUG_TRACE_FILE "constraint.out"  // uncomment to output a trace of
                                           // successes in adding constraints
//#define DEBUG_LOAD_TRACE_FILE "constraint.out" // uncomment to fake 
                                                 // constraint adds from file

//#define PRINT_STACK   // Uncomment to output stack traces
#define PRINT_FREQUENCY 100
// ^^ Says how many branch points should elapse between stack trace output

//#define ISAMP         // Uncomment, along with setting bNaiveSolver to implement ISAMP

//#define FAVOR_SMALL_CLAUSES // Uncomment to have small clauses favored for filtering

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Public Methods

bool VariableStruct::bFlush()
{
  bool bReturn = 0;
  list<VariableStruct *>::iterator i;
  
  for (i = _xCache.begin(); i != _xCache.end() && !bReturn; i++) {
    bReturn = (*i)->_bAddConstraint();
  }

  _xCache.clear();
  _iCacheSize = 0;

  return bReturn;
}

VariableStruct::~VariableStruct()
{
  if (pConstraint != 0 && bInSolver)
    xSolver.removeConstraint(*pConstraint);
}

void VariableStruct::vSetConstraint(ClConstraint * pcn)
{

  pConstraint = pcn;

  if (pcn)
    myLen = pConstraint->expression().terms().size();
  else 
    myLen = 0;
}

int VariableStruct::iConsSize()
{
  return myLen;
}

void VariableStruct::_vRemoveConstraint() 
{
  assert(pConstraint != 0);
  assert(eAssignment != 1);
  assert(bInSolver);

#ifndef NSTATS
  stats["Constraint deletes"].log(pSATSolver->iBranchDepth());
#endif

#ifndef NDEEPDEBUG
  cerr << "DELing constraint " << *pConstraint << endl;
#endif
  
#ifndef DEBUG_LOAD_TRACE_FILE

  xSolver.removeConstraint(*pConstraint);

#endif

  bInSolver = 0;
}

boolean VariableStruct::_bAddConstraint() 
{
  assert(pConstraint != 0);
  assert(eAssignment == 1);
  assert(!bInSolver);

#ifndef NSTATS
  stats["Constraint adds"].log(pSATSolver->iBranchDepth());
#endif
  
#ifdef DEBUG_LOAD_TRACE_FILE

  static ifstream in(DEBUG_LOAD_TRACE_FILE);

  if (!(in >> bInSolver)) {
    while (1) {
	 ;
    }
  }

#else

  // Note that the solver returns 1 when all's clear.
  // Note also that the constraint is not in the tableau unless
  // it can be successfully added!
  
#ifndef NDEEPDEBUG
  cerr << "ADDing constraint " << *pConstraint;
#endif
  
  try {
    xSolver.addConstraint(*pConstraint);
    bInSolver = true;
  }
  catch (const ExCLRequiredFailureWithExplanation & e) {
    // Gather the explanation
    xExplanation.vClear();
    ClConstraintSet::const_iterator i;
    for (i = e.explanation().begin(); i != e.explanation().end(); i++) {
	 assert(*i);
	 xExplanation.vAddVariable((VariableID)(*i)->Pv());
    }

    bInSolver = false;
  }
  catch (const ExCLRequiredFailure & e) {
    // No explanation available
    xExplanation.vClear();

    bInSolver = false;
  }

#endif

#ifndef NSTATS
  if (bInSolver)
    stats["Successful constraint adds"].log(pSATSolver->iBranchDepth());
#endif

#ifdef DEBUG_TRACE_FILE
  static ofstream out(DEBUG_TRACE_FILE);
  out << bInSolver << endl;
#endif
  
#ifndef NDEEPDEBUG
  if (bInSolver) {
    cerr << "  SUCCESS!" << endl;
  }
  else {
    cerr << "  FAILURE." << endl;
  }
#endif
  
  return !bInSolver;
}

SATSolver::SATSolver()
{
  vReset();
}

void SATSolver::vReset()
{
  // DOESN'T WORK!! YOU CAN ONLY USE A SOLVER ONCE RIGHT NOW!!!
  _lElapsedTime = 0;
  _lVariablesLabelled = 0;
  _iCurrentVariable = 0;
  _iBranchDepth = 0;
  _eCurrentID = NON_ID;
  _iVariableCount = 0;
  _eContradictionID = NON_ID;
  _bLogicContradiction = false;
  _bConsContradiction = false;

  _xUnitVariables0.vClear();
  _xUnitVariables1.vClear();
  _xUnlabeledSet.vClear();
}

boolean SATSolver::bGetValue(VariableID eVar) 
{ 
  assert(eVar >= 0 && eVar < _iVariableCount);
  return _aVariableStruct[eVar].eGetAssignment(); 
}

double SATSolver::dGetValue(int iVar) 
{
  assert(iVar >= 0 && iVar < SATInstance::xSATInstance.iRVariableCount);
  return SATInstance::xSATInstance.aVariables[iVar].value();
}

void SATSolver::_vBackOut()
{
  while (!VariableStruct::xSolver.ConstraintMap().empty()) {
    VariableStruct::xSolver.removeConstraint(*(*VariableStruct::xSolver.
									  ConstraintMap().begin()).first);
  }
}

boolean SATSolver::bFindSolution(boolean& rFailed_)
{
  // Returns 1 if problem is soluble, returns 0 if unsoluble. rFailed_
  // is set to 1 if the problem is too hard.

  //!!!VariableStruct::xSolver.reset();
  // Until reset works, need to back out of a previous solve attempt
  // gracefully
  _vBackOut();

  VariableStruct::pSATSolver = this;
  assert(VariableStruct::iCacheSize() == 0);
  VariableStruct::bSetAutoFlush(iCachePolicy == NO_CACHE);
  rFailed_ = 0;
  _iCurrentVariable = 0;
  _iBranchDepth = 0;
  _xUnitVariables0.vClear();
  _xUnitVariables1.vClear();
  _xBackupSet.vClear();
  _xNegatedBackupSet.vClear();
  _xPermaVariables.vClear();

  // Wolf - have to set up the solver not to solve all the time!
  VariableStruct::xSolver.setAutosolve(false);

  // Wolf - set explanations appropriately
  VariableStruct::xSolver.setExplaining(bExplanation);


  if (!_bInitializeVariableInformation())
    return 0; // problem is UNSAT due to node consistency

  // Here we do an initial unit propagation since initialization
  // can discover unit domained variables due to unary clauses.

  times(&buffer);
  _lElapsedTime = buffer.tms_utime + buffer.tms_stime;

  if (_bUnitPropagate()) {
    return 0;
  }

  boolean bReturn =  _bLoop(rFailed_);
  times(&buffer);
  _lElapsedTime = (buffer.tms_utime + buffer.tms_stime) - _lElapsedTime;
  VariableStruct::xSolver.solve();

  return bReturn;
}

void SATSolver::vOutput(ostream & out)
{
  int i;
  for (i=0; i<_iVariableCount; i++) {
    // Wolf - since this assumes unassigned vars are false, it doesn't
    //        run into trouble with unexpected constraints. I didn't
    //        change this. Just thought I'd make a note of it.
    if (_aVariableStruct[i].eGetAssignment())
	 out << i+1 << ' ';
    else out << "-" << i+1 << ' ';
  }
  out << "0" << endl;
  
  for (i = 0; i < SATInstance::xSATInstance.iRVariableCount; i++) {
    // Wolf - have to print out the values of the constrained variables:
    out << i+1 << " " << SATInstance::xSATInstance.aVariables[i].value() << endl;
  }
  out << "0" << endl;
}

boolean SATSolver::_bLoop(boolean& rFailed_)
{
  VariableSet xTryVars;
  int iLastBranchFlush = 0;
  while (_iCurrentVariable != _iVariableCount) {
    if (_iBranchDepth < iLastBranchFlush)
	 iLastBranchFlush = _iBranchDepth;

    stats["Solver loop"].log(iBranchDepth());
    assert(_xUnitVariables0.iVariableCount() == 0);
    assert(_xUnitVariables1.iVariableCount() == 0);

    // Here goes a branch:
    _iBranchDepth++;
    if (iCachePolicy == CACHE_BRANCH &&
	   _iBranchDepth - iLastBranchFlush > iCacheDepth) {
	 iLastBranchFlush = _iBranchDepth;
	 if (VariableStruct::bFlush()) {
	   _bConsContradiction = true;
	   if (_bBackup()) 
		return 0;
	   else
		continue;
	 }
    }

    VariableID eBestID;

    _vFindBestSet(xTryVars);

    boolean bZeroFirst = 0;

    if (bTimeLimit) {
	 times(&buffer);
	 if (((buffer.tms_utime + buffer.tms_stime) - _lElapsedTime) > iMaxTime) {
	   rFailed_ = 1;
	   return 0;
	 }
    }


    if (xTryVars.iVariableCount() == 1) {
      eBestID = xTryVars.eVariable(0);
    }
    else {
      eBestID = _eTryAndPickBest(xTryVars, bZeroFirst);
    }

    if (bUsePureLitElim) {
	 // Search for a pure literal.  This implementation of pure literal
	 // rule only checks to see if the variable selected as "best"
	 // appears as a pure literal. It does not check EVERY open
	 // variable for pure literal status.  The implementation exploits
	 // the fact that CBJ will backjump over the pure literal instead
	 // of branch on it given appropriate value ordering.

	 ClauseList* pClauseList = &_aVariableStruct[eBestID].xPositiveClauses;
	 boolean bAllSatisfied = 1;
	 for (int i=0; i < pClauseList->iClauseCount(); i++) {
	   if (!pClauseList->pClause(i)->bIsSatisfied()) {
		bAllSatisfied = 0;
		break;
	   }
	 }
	 if (bAllSatisfied) {
	   bZeroFirst = 1;
	 }
	 else {
	   pClauseList = &_aVariableStruct[eBestID].xNegativeClauses;
	   bAllSatisfied = 1;
	   for (int i=0; i < pClauseList->iClauseCount(); i++) {
		if (!pClauseList->pClause(i)->bIsSatisfied()) {
		  bAllSatisfied = 0;
		  break;
		}
	   }
	   if (bAllSatisfied) {
		bZeroFirst = 0;
	   }
	 }
    }

    // Special exception for constrained variables if the flag is set:
    // constrained vars always set to false first.
    if (bConstrainedVarsNegFirst && 
	   _aVariableStruct[eBestID].bIsConstrained()) {
	 bZeroFirst = 1;
    }

    if (bZeroFirst) {
      _xUnitVariables0.vAddVariableNoCheck(eBestID);
    }
    else {
      _xUnitVariables1.vAddVariableNoCheck(eBestID);
    }
    if (_bUnitPropagate()) {
      if (_bBackup()) {
	   return 0;
      }
    }

    if (_iCurrentVariable == _iVariableCount &&
	   iCachePolicy != NO_CACHE) {
	 iLastBranchFlush = _iBranchDepth;
	 if (VariableStruct::bFlush()) {
	   _bConsContradiction = true;
	   if (_bBackup())
		return 0;
	 }
    }
  }
  return 1;
}

VariableID SATSolver::_eTryAndPickBest(VariableSet& rTryVars_, boolean& bZeroFirst)
{
  int i; // Wolf - added this declaration plus removed the ones 
	    // in for statements to handle new scoping rules on loop variables.
  
  // Performs a "fast" unit propagation for both labels of each variable 
  // in the set and returns the variable that leads to the most binary 
  // clauses, or the first variable that leads to a contradiction.
  // note: it might be a good idea to skip constrained variables in
  // the fast unit prop.
  assert(_xUnitVariables0.iVariableCount() == 0);
  assert(_xUnitVariables1.iVariableCount() == 0);
  _xBackupSet.vClear(); // danger: Reusing this variable
  double iBestScore = -HUGE_VAL;
  double iScore;
  double iNewScore;
  for (i=0; i<rTryVars_.iVariableCount(); i++) {
    VariableID eID = rTryVars_.eVariable(i);
    assert(_aVariableStruct[eID].eGetAssignment() == NON_VALUE);

    // Compute "false" score
    _xUnitVariables0.vAddVariableNoCheck(eID);
    iScore = _iComputeScore();
    if (iScore == -1) {
      bZeroFirst = 1;
      return eID;  // A dead end was hit
    }

    if (_aVariableStruct[eID].bIsConstrained())
	 stats["Heuristic constraint"].log();

    // Compute "true" score
    _xUnitVariables1.vAddVariableNoCheck(eID);
    iNewScore = _iComputeScore(); 
    if (iNewScore == -1) {
      bZeroFirst = 0;
      return eID;
    }

    double iCombinedScore = _iCombineScores(iScore, iNewScore);
    if (_aVariableStruct[eID].bIsConstrained()) {
	 iCombinedScore += constraintPreferenceFactor;
	 iCombinedScore += _aVariableStruct[eID].iConsSize() *
	   constraintSizeFactor;
    }

    if (iCombinedScore >= iBestScore) {
      iBestScore = iCombinedScore;
      _xBackupSet.vAddVariable(eID);
      _aScore[eID] = iCombinedScore;
    }
  } // for (i=0;...
  rTryVars_.vClear();
  double factInv = (bestScoreFudgeFactor != 0) ? 1 / bestScoreFudgeFactor : 1000;
  double factor = (iBestScore >= 0) ? bestScoreFudgeFactor : factInv;
  double iAcceptableScore = iBestScore * factor;
  for (i=0; i<_xBackupSet.iVariableCount(); i++) {
    if (_aScore[_xBackupSet.eVariable(i)] >= iAcceptableScore) {
      rTryVars_.vAddVariable(_xBackupSet.eVariable(i));
    }
  }
  bZeroFirst = Random::iRandom(2);
  VariableID eReturn = 
    rTryVars_.eVariable(Random::iRandom(rTryVars_.iVariableCount()));
  return eReturn;
}


//////////////////////////////////////////////////////////////////////////////
// Protected Methods

//////////////////////////////////////////////////////////////////////////////
// Private Methods
boolean SATSolver::_bInitializeVariableInformation()
{
  int i; // Wolf - added this declaration plus removed the ones in for 
         // statements to handle new scoping rules on loop variables.
  
  // Return 0 if problem is UNSAT due to unary clauses (node-consistency)
  _eCurrentID = NON_ID;
  _iVariableCount = SATInstance::xSATInstance.iVariableCount;
  for (i=0; i< _iVariableCount; i++) {
    Clause::a2ClauseCount[i*2] = 0;
    Clause::a2ClauseCount[i*2+1] = 0;

    _aVariableStruct[i].vFastUnsetAssignment(false);
    _aVariableStruct[i].pDeleteClause = 0;
    _aVariableStruct[i].pFilterClause0 = 0;
    _aVariableStruct[i].pFilterClause1 = 0;
    _aVariableStruct[i].
	 vSetConstraint(SATInstance::xSATInstance.aConstraints[i]);
    _xUnlabeledSet.vAddVariable(i);
  }

  cerr << "Finished ADDING!" << endl;

  for (i=0; i<SATInstance::xSATInstance.iClauseCount(); i++) {
    Clause* pClause = SATInstance::xSATInstance.pClause(i);
    if (pClause->iVariableCount() == 2) {
      _vInitializeBinaryClause(pClause);
    }
    else if (pClause->iVariableCount() == 1) {
      if (_bInitializeUnaryClause(pClause) == 0)
	   return 0;
    }
    else {
      _vInitializeLargeClause(pClause);
    }
  }

  cerr << "Finished init!" << endl;

  return 1;
}

inline void SATSolver::_vInitializeLargeClause(Clause* pClause_)
{
  for (int j=0; j<pClause_->iPermaCount(); j++) {
    if (pClause_->bIsNegated(j))
      _aVariableStruct[pClause_->eConstrainedVariable(j)].
	   xNegativeClauses.vAddClause(pClause_);
    else
      _aVariableStruct[pClause_->eConstrainedVariable(j)].
	   xPositiveClauses.vAddClause(pClause_);
  }
}

inline void SATSolver::_vInitializeBinaryClause(Clause* pClause_)
{
  assert(pClause_->iVariableCount() == 2);
  for (int j=0; j<2; j++) {
    if (pClause_->bIsNegated(j))
      _aVariableStruct[pClause_->eConstrainedVariable(j)].xNegativeClauses.vAddClause(pClause_);
    else
      _aVariableStruct[pClause_->eConstrainedVariable(j)].xPositiveClauses.vAddClause(pClause_);
    Clause::a2ClauseCount[pClause_->iLiteralID(j)]++;
  }
}

boolean SATSolver::_bInitializeUnaryClause(Clause* pClause_)
{
  // returns 0 if problem is determined to be UNSAT 
  assert(pClause_->iVariableCount() == 1);
  VariableID eConstrainedVariable = pClause_->eConstrainedVariable(0);
  if (pClause_->bIsNegated(0)) {
    if (_aVariableStruct[eConstrainedVariable].pFilterClause0)
      return 0; // problem is UNSAT
    if (_aVariableStruct[eConstrainedVariable].pFilterClause1 == 0) {
      _xUnitVariables0.vAddVariable(eConstrainedVariable);
      _aVariableStruct[eConstrainedVariable].pFilterClause1 = pClause_;
    }
  }
  else {
    if (_aVariableStruct[eConstrainedVariable].pFilterClause1)
      return 0; // problem is UNSAT
    if (_aVariableStruct[eConstrainedVariable].pFilterClause0 == 0) {
      _xUnitVariables1.vAddVariable(eConstrainedVariable);
      _aVariableStruct[eConstrainedVariable].pFilterClause0 = pClause_;
    }
  }
  return 1;
}

void SATSolver::_vFillCandidateSet(double & iBestScore, int & numCandidates)
{
  double cutoff = iBestScore * scoreCutoffFactor;
  numCandidates = 0;
  for (int j = 0; j < _xUnlabeledSet.iVariableCount(); j++) {
    VariableID i = _xUnlabeledSet.eVariable(j);

    _aScore[i] = _iCombineScores(Clause::a2ClauseCount[i*2], 
						   Clause::a2ClauseCount[i*2+1]);

    // Can get this almost for free now that lengths are cached.
    _aScore[i] += _aVariableStruct[i].iConsSize() * constraintSizeFactor;

//	 if (_aVariableStruct[i].bIsConstrained()) {
//	   _aScore[i] += constraintPreferenceFactor;
	   // Takes _too_long_! Biggest portion of time
//	 }
    if (_aScore[i] >= cutoff) {
	 _aCandidates[numCandidates++] = i;
	 if (_aScore[i] > iBestScore) {
	   iBestScore = _aScore[i];
	   cutoff = iBestScore * scoreCutoffFactor;
	 }
    }
  }
}

// Wolf - I'm modifying this to allow a multiplicative preference factor
// for constrained vars.
void SATSolver::_vFindBestSet(VariableSet& xTryVars)
{
  xTryVars.vClear();
#ifdef ISAMP
  // select random unlabeled variable
  VariableID eID;
  do {
    eID = Random::iRandom(_iVariableCount);
  } while (_aVariableStruct[eID].eGetAssignment() != NON_VALUE);
  xTryVars.vAddVariable(eID);
  return;
#endif ISAMP
  double iBestScore = 1.0;
  int iNumValuable;
  _xBackupSet.vClear(); // danger: reusing this variable
  _vFillCandidateSet(iBestScore, iNumValuable);

  if (iNumValuable) {
    double factInv = (scoreCutoffFactor != 0) ? 1 / scoreCutoffFactor : 1000;
    double factor = (iBestScore >= 0) ? scoreCutoffFactor : factInv;
    double iAcceptableScore = iBestScore * factor;
    for (int i=0; i < iNumValuable; i++) {
      VariableID eID = _aCandidates[i];
      if (_aScore[eID] >= iAcceptableScore)
	   xTryVars.vAddVariable(eID);
    }
    while (xTryVars.iVariableCount() > MAX_TO_TRY) {
      xTryVars.vRemoveVariable(xTryVars.eVariable(Random::iRandom(xTryVars.iVariableCount())));
    }
  }
  else {
    // All the variables are worthless so select a random unlabeled variable
    //    cout << "Selecting a worthless variable\n"; 
    VariableID eID;
    do {
      eID = Random::iRandom(_iVariableCount);
    } while (_aVariableStruct[eID].eGetAssignment() != NON_VALUE);
    xTryVars.vAddVariable(eID);
  }
}

double SATSolver::_iComputeScore()
{
  // Before calling, ensure that there is exactly one unit variable.
  // On exit, ensures that there are no unit variables.
  // Returns the number of variables labeled by performing unit propagation, 
  // or -1 if a contradiction is derived.
  assert(_xUnitVariables0.iVariableCount() == 1 || 
	    _xUnitVariables1.iVariableCount() == 1);
  assert(_xUnitVariables0.iVariableCount() == 0 || 
	    _xUnitVariables1.iVariableCount() == 0);

  int iRememberVariableCount = _iCurrentVariable;
  VariableID eBackupID = _eCurrentID; 
  boolean bContradiction = _bFastUnitPropagate();
  if (bContradiction) {
    _vFastBackup(eBackupID);
    _xUnitVariables0.vClear();
    _xUnitVariables1.vClear();
    return -1;  
  }
  int iReturn = (_iCurrentVariable - iRememberVariableCount);
  _vFastBackup(eBackupID);
  return iReturn;
}

void SATSolver::_vFastBackup(VariableID eBackupID_)
{
  Clause** aClauseList1;
  Clause** aClauseList2;
  int iTo1, iTo2, iTo;
  register int i;
  while (1) { 
    if (_aVariableStruct[_eCurrentID].eGetAssignment() == 1) {
      aClauseList1 = _aVariableStruct[_eCurrentID].xNegativeClauses._aClause;
      iTo1 = _aVariableStruct[_eCurrentID].xNegativeClauses.iClauseCount();
      aClauseList2 = _aVariableStruct[_eCurrentID].xPositiveClauses._aClause;
      iTo2 = _aVariableStruct[_eCurrentID].xPositiveClauses.iClauseCount();
    }
    else {
      aClauseList1 = _aVariableStruct[_eCurrentID].xPositiveClauses._aClause;
      iTo1 = _aVariableStruct[_eCurrentID].xPositiveClauses.iClauseCount();
      aClauseList2 = _aVariableStruct[_eCurrentID].xNegativeClauses._aClause;
      iTo2 = _aVariableStruct[_eCurrentID].xNegativeClauses.iClauseCount();
    }

    if (iTo1 > iTo2) {
      for (i = 0; i < iTo2; i++) {
	   aClauseList1[i]->vFastUnfilter();
	   aClauseList2[i]->vFastMakeUnsatisfied();
      }
      for (; i < iTo1; i++) {
	   aClauseList1[i]->vFastUnfilter();
      }
    }
    else {
      for (i = 0; i < iTo1; i++) {
	   aClauseList1[i]->vFastUnfilter();
	   aClauseList2[i]->vFastMakeUnsatisfied();
      }      
      for (; i < iTo2; i++) {
	   aClauseList2[i]->vFastMakeUnsatisfied();
      }      
    }

    _aVariableStruct[_eCurrentID].vFastUnsetAssignment(fastUnitPropConstrained);
    _iCurrentVariable--;

    // special code only needed for labeling first variable..
    if (_iCurrentVariable == 0) {
      _eCurrentID = -1;
      break;
    }

    _eCurrentID = _aPositionToID[_iCurrentVariable-1];
    if (_eCurrentID == eBackupID_)
      break;
  } //while(1);
}

boolean SATSolver::_bFastUnitPropagate()
{
  // Return 1 if unit propagation leads to a contradiction.
  // Does not back up, does not learn.

  int iTo1, iTo2, iInnerTo; 
  register int i;
  int j;
  Clause** aClauseList1;
  Clause** aClauseList2;
  Clause* pWorkClause;
  boolean bReturn;
  VariableID eID;
  while(1) {
    // Wolf - moved this statement up and added bReturn checks on the labeling
    //        to handle possible snags on constraints
    bReturn = 0;

    if (_xUnitVariables1.iVariableCount()) {
      bReturn = _bFastLabelVariable(_xUnitVariables1.eVariable(0), 1,
							 fastUnitPropConstrained);
      _xUnitVariables1.vRemoveVariable(_eCurrentID);
      aClauseList1 = _aVariableStruct[_eCurrentID].xPositiveClauses._aClause;
      iTo1 = _aVariableStruct[_eCurrentID].xPositiveClauses.iClauseCount();
      aClauseList2 = _aVariableStruct[_eCurrentID].xNegativeClauses._aClause;
      iTo2 = _aVariableStruct[_eCurrentID].xNegativeClauses.iClauseCount();
    }
    else if (_xUnitVariables0.iVariableCount()) {
      bReturn = _bFastLabelVariable(_xUnitVariables0.eVariable(0), 0,
							 fastUnitPropConstrained);
      _xUnitVariables0.vRemoveVariable(_eCurrentID);
      aClauseList1 = _aVariableStruct[_eCurrentID].xNegativeClauses._aClause;
      iTo1 = _aVariableStruct[_eCurrentID].xNegativeClauses.iClauseCount();
      aClauseList2 = _aVariableStruct[_eCurrentID].xPositiveClauses._aClause;
      iTo2 = _aVariableStruct[_eCurrentID].xPositiveClauses.iClauseCount();
    }
    else return 0;

    if (bReturn)
	 stats["Heuristic contradiction"].log();

    for (i = 0; i < iTo1; i++) {
      aClauseList1[i]->vFastMakeSatisfied();
    }

    for (i = 0; i < iTo2; i++) {
      pWorkClause = aClauseList2[i];
      if (pWorkClause->bFastFilter()) {
	   // Find the variable it filters.
	   iInnerTo = pWorkClause->iVariableCount();
	   for (j = 0; j < iInnerTo; j++) {
		eID = pWorkClause->eConstrainedVariable(j);
		if (_aVariableStruct[eID].eGetAssignment() == NON_VALUE) {
		  // we've found the filtering variable

// Using new technique to handle not propagating constraints
//		  // if the variable is constrained and we're not
//		  // traversing through constraints, skip this clause:
//		  if (!fastUnitPropConstrained && 
//			 _aVariableStruct[eID].bIsConstrained()) {
//		    break;
//		  }

		  if (pWorkClause->bIsNegated(j)) {
		    _xUnitVariables0.vAddVariable(eID);
		    if (_xUnitVariables1.bHasVariable(eID)) {
			 // Can't simply return immediately because it will 
			 // confuse the state undoing code
			 bReturn = 1;
		    }
		  }
		  else  {
		    _xUnitVariables1.vAddVariable(eID);
		    if (_xUnitVariables0.bHasVariable(eID)) {
			 // Can't simply return immediately because it will 
			 // confuse the state undoing code
			 bReturn = 1;
		    }
		  }
		  break;
		}
	   }
      }
    } // for i

    if (bReturn == 1) {
      return 1;
    }

    if (iCachePolicy == CACHE_SIZE && 
	   VariableStruct::iCacheSize() > iCacheDepth) {
	 if (VariableStruct::bFlush())
	   return 1;
    }
  } //while(1);
}

boolean SATSolver::_bUnitPropagate()
{
  // Return 1 if unit propagation leads to a contradiction.
  // Does not back up, does not learn.
  // Simply leaves the list of contradicting variables intact.

  _bLogicContradiction = false;
  _bConsContradiction = false;

  while(1) {
    if (_xUnitVariables1.iVariableCount()) {
      _bConsContradiction = _bLabelVariable(_xUnitVariables1.eVariable(0), 1);
      _xUnitVariables1.vRemoveVariable(_eCurrentID);
      _vSatisfyWithClauseList(_aVariableStruct[_eCurrentID].xPositiveClauses);
	 _bLogicContradiction = 
	   _bFilterWithClauseList(_aVariableStruct[_eCurrentID].xNegativeClauses);

	 if (_bConsContradiction || _bLogicContradiction)
	   return 1;
    }
    else if (_xUnitVariables0.iVariableCount()) {
      _bConsContradiction = _bLabelVariable(_xUnitVariables0.eVariable(0), 0);
      _xUnitVariables0.vRemoveVariable(_eCurrentID);
      _vSatisfyWithClauseList(_aVariableStruct[_eCurrentID].xNegativeClauses);
      _bLogicContradiction = 
	   _bFilterWithClauseList(_aVariableStruct[_eCurrentID].xPositiveClauses);

	 if (_bConsContradiction || _bLogicContradiction)
	   return 1;
    }
    else return 0;

    if (iCachePolicy == CACHE_SIZE && 
	   VariableStruct::iCacheSize() > iCacheDepth) {
	 if (VariableStruct::bFlush()) {
	   _bConsContradiction = 1;
	   return 1;
	 }
    }
  } // while(1);
}


void SATSolver::_vUnfilterWithClauseList(ClauseList& rClauseList_)
{
  int iTo =  rClauseList_.iClauseCount();
  for (int i = 0; i < iTo; i++) {
    rClauseList_.pClause(i)->vUnfilter();
  }
}

boolean SATSolver::_bFilterWithClauseList(ClauseList& rClauseList_)
{
  // return 1 if contradiction encountered.

  int iTo =  rClauseList_.iClauseCount();
  boolean bReturn = 0;
  Clause* pWorkClause;
  for (int i = 0; i < iTo; i++) {
    pWorkClause = rClauseList_.pClause(i);
    if (pWorkClause->bFilter()) {
      // Find the variable it filters.
      for (int j = 0; j < pWorkClause->iVariableCount(); j++) {
	   VariableID eID = pWorkClause->eConstrainedVariable(j);
	   if (_aVariableStruct[eID].eGetAssignment() == NON_VALUE) {
		// we've found the filtering variable
		if (pWorkClause->bIsNegated(j)) {

		  if (_aVariableStruct[eID].pFilterClause1) {
#ifdef FAVOR_SMALL_CLAUSES
		    if (_aVariableStruct[eID].pFilterClause1->iVariableCount() >
			   pWorkClause->iVariableCount()) {
			 _aVariableStruct[eID].pFilterClause1 = pWorkClause;
		    }
		    else
			 if (_aVariableStruct[eID].pFilterClause1->iVariableCount() ==
				pWorkClause->iVariableCount()) 
#endif
			   { 
				if (Random::bRandom()) 
				  _aVariableStruct[eID].pFilterClause1 = pWorkClause;
			   }
		  }
		  else {
		    _aVariableStruct[eID].pFilterClause1 = pWorkClause;
		    _xUnitVariables0.vAddVariable(eID);
		    if (_xUnitVariables1.bHasVariable(eID)) {
			 // Can't simply return immediately because it will 
			 // confuse the state undoing code
			 _eContradictionID = eID;
			 bReturn = 1;
		    }
		  }
		}
		else  {
		  if (_aVariableStruct[eID].pFilterClause0) {
#ifdef FAVOR_SMALL_CLAUSES
		    if (_aVariableStruct[eID].pFilterClause0->iVariableCount() >
			   pWorkClause->iVariableCount()) 
			 _aVariableStruct[eID].pFilterClause0 = pWorkClause;
		    else 
			 if (_aVariableStruct[eID].pFilterClause0->iVariableCount() ==
				pWorkClause->iVariableCount()) 
#endif
			   { 
				if (Random::bRandom())
				  _aVariableStruct[eID].pFilterClause0 = pWorkClause;
			   }
		  }
		  else {
		    _aVariableStruct[eID].pFilterClause0 = pWorkClause;
		    _xUnitVariables1.vAddVariable(eID);
		    if (_xUnitVariables0.bHasVariable(eID)) {
			 // Can't simply return immediately because it will 
			 // confuse the state undoing code
			 _eContradictionID = eID;
			 bReturn = 1;
		    }
		  }
		  break;
		}
	   }
      }
    }
  }
  return bReturn;
}

void SATSolver::_vSatisfyWithClauseList(ClauseList& rClauseList_)
{
  int iTo =  rClauseList_.iClauseCount();
  for (int i = 0; i < iTo; i++) {
    rClauseList_.pClause(i)->vMakeSatisfied();
  }
}


void SATSolver::_vCreateBackupClauseFromContradictions()
{
  // Create a new clause representing the nogood derived from the
  // current contradiction on _eContradictionID.

  // For now, I'm choosing to ignore logical contradictions if there is
  // a constraint contradiction. Perhaps I can do both? !!!
  if (_bConsContradiction) {
    _vCreateBackupClauseFromConstraintContradictions();
    _bLogicContradiction = false;
  }
  else if (_bLogicContradiction)
    _vCreateBackupClauseFromLogicalContradictions();
}

void SATSolver::_vCreateBackupClauseFromConstraintContradictions()
{
  int i;
  VariableID eID;

  for (i=0; i<_xUnitVariables0.iVariableCount(); i++) {
    _aVariableStruct[_xUnitVariables0.eVariable(i)].pFilterClause1 = 0;
  }
  _xUnitVariables0.vClear();
  for (i=0; i<_xUnitVariables1.iVariableCount(); i++) {
    _aVariableStruct[_xUnitVariables1.eVariable(i)].pFilterClause0 = 0;
  }
  _xUnitVariables1.vClear();

  _xBackupSet.vClear();
  _xNegatedBackupSet.vClear();
  
  _xBackupSet.vUnion(VariableStruct::xExplanation);
  _xNegatedBackupSet.vUnion(VariableStruct::xExplanation);

  // if these sets are empty, it means no explanation was given, and
  // we must assume that any active constraint could have caused the 
  // failure.
  if (_xBackupSet.iVariableCount() == 0) {
    // Add the negations of exactly those variables which may be causing the
    // conflict to the backup set.
    for (i = 0; i < _iCurrentVariable; i++) {
	 eID = _aPositionToID[i];
	 if (_aVariableStruct[eID].bIsConstrained() &&
		_aVariableStruct[eID].eGetAssignment() == 1) {
	   _xBackupSet.vAddVariableNoCheck(eID);
	   _xNegatedBackupSet.vAddVariableNoCheck(eID);
	 }
    }
  }

  stats["Constraint backup clause"].log(_xBackupSet.iVariableCount());
}

void SATSolver::_vCreateBackupClauseFromLogicalContradictions()
{
  int i; // Wolf - added this declaration plus removed the ones in 
         // for statements to handle new scoping rules on loop variables.
  
  Clause* pFailClause1 = _aVariableStruct[_eContradictionID].pFilterClause1;
  Clause* pFailClause2 = _aVariableStruct[_eContradictionID].pFilterClause0;

  assert(_aVariableStruct[_eContradictionID].pFilterClause1);
  assert(_aVariableStruct[_eContradictionID].pFilterClause0);

  for (i=0; i<_xUnitVariables0.iVariableCount(); i++) {
    _aVariableStruct[_xUnitVariables0.eVariable(i)].pFilterClause1 = 0;
  }
  _xUnitVariables0.vClear();
  for (i=0; i<_xUnitVariables1.iVariableCount(); i++) {
    _aVariableStruct[_xUnitVariables1.eVariable(i)].pFilterClause0 = 0;
  }
  _xUnitVariables1.vClear();

  _xBackupSet.vClear();
  _xNegatedBackupSet.vClear();
  for (i=0; i<pFailClause1->iVariableCount(); i++) {
    _xBackupSet.vAddVariableNoCheck(pFailClause1->eConstrainedVariable(i));
    if (pFailClause1->bIsNegated(i)) {
      _xNegatedBackupSet.
	   vAddVariableNoCheck(pFailClause1->eConstrainedVariable(i));
    }
  }
  for (i=0; i<pFailClause2->iVariableCount(); i++) {
    if (_xBackupSet.bAddVariable(pFailClause2->eConstrainedVariable(i)) && 
	   pFailClause2->bIsNegated(i)) {
      _xNegatedBackupSet.
	   vAddVariableNoCheck(pFailClause2->eConstrainedVariable(i));
    }
  }
  _xBackupSet.vRemoveVariable(_eContradictionID);
  _xNegatedBackupSet.vRemoveVariableCheck(_eContradictionID);
}

inline void SATSolver::_vCreateNewBackupClause()
{
  Clause* pFailClause2;
  int iFailCount1 = _xBackupSet.iVariableCount();
  if (_aVariableStruct[_eCurrentID].pFilterClause0 == 0) {
    pFailClause2 = _aVariableStruct[_eCurrentID].pFilterClause1;
  }
  else {
    pFailClause2 = _aVariableStruct[_eCurrentID].pFilterClause0;
  }
  for (int i=0; i<pFailClause2->iVariableCount(); i++) {
    VariableID eWorkID = pFailClause2->eConstrainedVariable(i);
    if (_xBackupSet.bAddVariable(eWorkID) && pFailClause2->bIsNegated(i)) {
      _xNegatedBackupSet.vAddVariableNoCheck(eWorkID);
    }
  }  
  _xBackupSet.vRemoveVariable(_eCurrentID);
  _xNegatedBackupSet.vRemoveVariableCheck(_eCurrentID);
}

inline void SATSolver::_vLearn()
{
  int _iBackups = _xBackupSet.iVariableCount();

#ifndef NSTATS
  stats["Learn clause"].log(_xBackupSet.iVariableCount());
#endif

  if (_iBackups <= iLogicMaxLearnSize) {
    if (_xBackupSet.iVariableCount()==1) {
      // Unary Clause! This implementation can't really learn them yet, 
	 // but they arise pretty rarely so learning them doesn't really 
	 // seem to help much.
    }
    else {
      Clause* pLearn = new Clause(_xBackupSet, _xNegatedBackupSet);
	 pLearn->vPrepareForUse();
	 _vInitializeLargeClause(pLearn);
	 //pLearn->vOutput();
    }  
  }
  else if (bRelevance) {
    // Wolf - based on the new explanation based learning, I'm
    //        axing separate logic/constraint learn lengths.
    int i = _iCurrentVariable - 1;

    _xPermaVariables.vClear();
    int iBranches = 0;
    VariableID eResetID;

    int iPermaCount = 0;
    
    while (1) {
	 VariableID eID = _aPositionToID[i];
	 assert(_aVariableStruct[eID].eGetAssignment() !=  NON_VALUE);
	 if (_xBackupSet.bHasVariable(eID)) {
	   if (iPermaCount >= iLogicMaxLearnSize) {
		eResetID = eID;
		break;
	   }
	   if (_aVariableStruct[eID].pFilterClause0 == 0 && 
		  _aVariableStruct[eID].pFilterClause1 == 0) 
		iBranches++;
	   _xPermaVariables.vAddVariableNoCheck(eID);
	   iPermaCount++;
	 }
	 i--;
	 
	 // i shouldn't be less than zero since everything in the backup clause
	 // is somewhere in the positiontoID list, and the backup clause is
	 // longer than one or the other of the learn lengths.
	 assert(i>=0);
    } // while
    assert(_xPermaVariables.iVariableCount() == iLogicMaxLearnSize);

    // If there's no branches along the way, why bother learning? It will
    // just become irrelevant in the ensuing backup.
    if (iBranches > 0) {
      Clause* pLearn = new Clause(_xBackupSet, _xNegatedBackupSet, 
						    _xPermaVariables);
      _vInitializeLargeClause(pLearn);
	 //cout << "Relclause ";
	 //pLearn->vOutput();
      _aVariableStruct[eResetID].xDeleteList.vAddClause(pLearn);
    }
  } // else 
}

inline void SATSolver::_vDeleteClauses(ClauseList& rClauseList_)
{
  // Delete these clauses from the problem 
  int iTo = rClauseList_.iClauseCount();
  for (int i=0; i< iTo; i++) {
    Clause* pDeleteMe = rClauseList_.pClause(i);
    for (int j=0; j<pDeleteMe->iPermaCount(); j++) {
      if (pDeleteMe->bIsNegated(j))
	   _aVariableStruct[pDeleteMe->eConstrainedVariable(j)].
		xNegativeClauses.vDeleteClause(pDeleteMe);
      else _aVariableStruct[pDeleteMe->eConstrainedVariable(j)].
		   xPositiveClauses.vDeleteClause(pDeleteMe);
      assert(_aVariableStruct[pDeleteMe->eConstrainedVariable(j)].
	 	   eGetAssignment() == NON_VALUE);
    }
    delete pDeleteMe;
  }
  rClauseList_.vClear();
}

inline boolean SATSolver::_bFastLabelVariable(VariableID eID_, 
									 DomainValue lWhich_,
									 bool bConstrain)
{
  // Label a variable and make it current.
  _aPositionToID[_iCurrentVariable++] = eID_;
  _eCurrentID = eID_;
  return _aVariableStruct[eID_].bFastSetAssignment(lWhich_, bConstrain);
}

inline boolean SATSolver::_bLabelVariable(VariableID eID_, DomainValue lWhich_)
{
  // Label a variable and make it current.
  stats["Labelled variables"].log(iBranchDepth());
  _aPositionToID[_iCurrentVariable++] = eID_;
  _eCurrentID = eID_;
  _xUnlabeledSet.vRemoveVariableCheck(eID_);
  return _aVariableStruct[eID_].bSetAssignment(lWhich_);
}


boolean SATSolver::_bBackup()
{
  if (bNaiveSolver)
    return _bNaiveBackup();
  else
    return _bLearnBackup();
}

boolean SATSolver::_bNaiveBackup()
{
  int k; // Wolf - added this declaration plus removed the ones in
         // for statements to handle new scoping rules on loop variables.
  
  // Returns 1 if there are no solutions.
start:
  for (k=0; k<_xUnitVariables0.iVariableCount(); k++) {
    _aVariableStruct[_xUnitVariables0.eVariable(k)].pFilterClause1 = 0;
  }
  _xUnitVariables0.vClear();
  for (k=0; k<_xUnitVariables1.iVariableCount(); k++) {
    _aVariableStruct[_xUnitVariables1.eVariable(k)].pFilterClause0 = 0;
  }
  _xUnitVariables1.vClear();

  Clause** aClause;
  Clause** aClause0;
  int iTo, iTo0;
  register int i;
  do { // while (1)
    if (_aVariableStruct[_eCurrentID].eGetAssignment() == 1) {
      iTo0 = _aVariableStruct[_eCurrentID].xNegativeClauses.iClauseCount();
      aClause0 = _aVariableStruct[_eCurrentID].xNegativeClauses._aClause;
      iTo = _aVariableStruct[_eCurrentID].xPositiveClauses.iClauseCount();
      aClause = _aVariableStruct[_eCurrentID].xPositiveClauses._aClause;
    }
    else {
      iTo0 = _aVariableStruct[_eCurrentID].xPositiveClauses.iClauseCount();
      aClause0 = _aVariableStruct[_eCurrentID].xPositiveClauses._aClause;
      iTo = _aVariableStruct[_eCurrentID].xNegativeClauses.iClauseCount();
      aClause = _aVariableStruct[_eCurrentID].xNegativeClauses._aClause;
    }
    for (i = 0; i < iTo0; i++) {
      aClause0[i]->vUnfilter();
    }
    for (i = 0; i < iTo; i++) {
      aClause[i]->vMakeUnsatisfied();
    }
#ifndef ISAMP
    if (_aVariableStruct[_eCurrentID].pFilterClause0 == 0 &&
	   _aVariableStruct[_eCurrentID].pFilterClause1 == 0) {
      // we have backed up to a branch
	 stats["Branch points"].log(iGetDepth());
#ifdef PRINT_STACK
	 if (stats["Branch points"].count() % PRINT_FREQUENCY == 1) {
	   int iNoBranch = 0;
	   cout << '>';
	   for (int k=0; k<_iCurrentVariable; k++) {
		VariableID eWorkID = _aPositionToID[k];
		if (_aVariableStruct[eWorkID].pFilterClause0 == 0 &&
		    _aVariableStruct[eWorkID].pFilterClause1 == 0) {
		  cout << iNoBranch << ' ';
		  iNoBranch = 0;
		}
		else {
		  iNoBranch++;
		}
	   }
	   cout << endl;
	 }
#endif PRINT_STACK
	 if (_aVariableStruct[_eCurrentID].eGetAssignment() == 0) {
	   // This code is for branch variables initially labeled with 0
	   _aVariableStruct[_eCurrentID].pFilterClause0 = (Clause*) 1;
	   _xUnitVariables1.vAddVariable(_eCurrentID);
      }
      else {
	   _aVariableStruct[_eCurrentID].pFilterClause1 = (Clause*) 1;
	   _xUnitVariables0.vAddVariable(_eCurrentID);
      }
      _iCurrentVariable--;
      if (_bUnitPropagate()) {
	   goto start;
      }
      return 0;
    }
    else {
	 if (_aVariableStruct[_eCurrentID].pFilterClause0 == (Clause*)1 &&
		_aVariableStruct[_eCurrentID].pFilterClause1 == (Clause*)1) {
	   // we are backing up past a branch
	   _iBranchDepth--;
	 }

      if (_iCurrentVariable == 1)
	   return 1;
    }
#endif ISAMP
    _xUnlabeledSet.vAddVariable(_eCurrentID);
    _aVariableStruct[_eCurrentID].vUnsetAssignment();
    _aVariableStruct[_eCurrentID].pFilterClause1 = 0;
    _aVariableStruct[_eCurrentID].pFilterClause0 = 0;
    _iCurrentVariable--;

    // special code only needed for labeling first variable...
    if (_iCurrentVariable == 0) {
      _eCurrentID = -1;
      return 0;
    }
    assert(_iCurrentVariable != 0);
    _eCurrentID = _aPositionToID[_iCurrentVariable-1];
  } while(1);
}

boolean SATSolver::_bLearnBackup()
{
  // Returns 1 if there are no solutions.
start:
  _vCreateBackupClauseFromContradictions();
  if (bLearn)
    _vLearn();

  Clause** aClause;
  Clause** aClause0;
  int iTo, iTo0;
  register int i;
  do { // while (1)
    if (_aVariableStruct[_eCurrentID].eGetAssignment() == 1) {
      iTo0 = _aVariableStruct[_eCurrentID].xNegativeClauses.iClauseCount();
      aClause0 = _aVariableStruct[_eCurrentID].xNegativeClauses._aClause;
      iTo = _aVariableStruct[_eCurrentID].xPositiveClauses.iClauseCount();
      aClause = _aVariableStruct[_eCurrentID].xPositiveClauses._aClause;
    }
    else {
      iTo0 = _aVariableStruct[_eCurrentID].xPositiveClauses.iClauseCount();
      aClause0 = _aVariableStruct[_eCurrentID].xPositiveClauses._aClause;
      iTo = _aVariableStruct[_eCurrentID].xNegativeClauses.iClauseCount();
      aClause = _aVariableStruct[_eCurrentID].xNegativeClauses._aClause;
    }
    for (i = 0; i < iTo0; i++) {
      aClause0[i]->vUnfilter();
    }
    for (i = 0; i < iTo; i++) {
      aClause[i]->vMakeUnsatisfied();
    }
    if (_xBackupSet.bHasVariable(_eCurrentID)) {
      if (_aVariableStruct[_eCurrentID].pFilterClause0 == 0 &&
		_aVariableStruct[_eCurrentID].pFilterClause1 == 0) {
	   _vDeleteClauses(_aVariableStruct[_eCurrentID].xDeleteList);
	   // we have backed up to a branch
	   stats["Branch points"].log(iGetDepth());
#ifdef PRINT_STACK
	   if (stats["Branch points"].count() % PRINT_FREQUENCY == 1) {
		int iNoBranch = 0;
		cout << '>';
		for (int k=0; k<_iCurrentVariable; k++) {
		  VariableID eWorkID = _aPositionToID[k];
		  if (_aVariableStruct[eWorkID].pFilterClause0 == 0 &&
			 _aVariableStruct[eWorkID].pFilterClause1 == 0) {
		    cout << iNoBranch << ' ';
		    iNoBranch = 0;
		  }
		  else {
		    iNoBranch++;
		  }
		}
		cout << endl;
	   }
#endif // PRINT_STACK
	   if (_aVariableStruct[_eCurrentID].eGetAssignment() == 0) {
		// This code is for branch variables initially labeled with 0
		_aVariableStruct[_eCurrentID].pFilterClause0 = 
		  new Clause(_xBackupSet, _xNegatedBackupSet);
		_aVariableStruct[_eCurrentID].pDeleteClause = 
		  _aVariableStruct[_eCurrentID].pFilterClause0;
		_xUnitVariables1.vAddVariable(_eCurrentID);
	   }
	   else {
		_aVariableStruct[_eCurrentID].pFilterClause1 = 
		  new Clause(_xBackupSet, _xNegatedBackupSet);
		_aVariableStruct[_eCurrentID].pDeleteClause = 
		  _aVariableStruct[_eCurrentID].pFilterClause1;
		_xUnitVariables0.vAddVariable(_eCurrentID);
	   }
	   _iCurrentVariable--;
	   if (_bUnitPropagate()) {
		goto start;
	   }
	   return 0;
      }
      else {
	   _vCreateNewBackupClause();
	   if (_xBackupSet.iVariableCount() == 0)
		return 1;
	   if (bLearn)
		_vLearn();
      }
    }
    // Check for a former branch point:
    if ((_aVariableStruct[_eCurrentID].pFilterClause0 == 0 &&
	    _aVariableStruct[_eCurrentID].pFilterClause1 == 0) ||
	   _aVariableStruct[_eCurrentID].pDeleteClause != 0) {
	 _iBranchDepth--;

	 if (_aVariableStruct[_eCurrentID].pDeleteClause != 0) {
	   delete _aVariableStruct[_eCurrentID].pDeleteClause;
	   _aVariableStruct[_eCurrentID].pDeleteClause = 0;
	 }
    }
    _xUnlabeledSet.vAddVariable(_eCurrentID);
    _aVariableStruct[_eCurrentID].vUnsetAssignment();
    _aVariableStruct[_eCurrentID].pFilterClause1 = 0;
    _aVariableStruct[_eCurrentID].pFilterClause0 = 0;
    _vDeleteClauses(_aVariableStruct[_eCurrentID].xDeleteList);
    _iCurrentVariable--;
    assert(_iCurrentVariable != 0);
    _eCurrentID = _aPositionToID[_iCurrentVariable-1];
  } while(1);
}

