#ifndef SATSolver_h
#define SATSolver_h

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// SATSolver

/////////////////
// OS Includes
#include <sys/types.h>
#include <sys/times.h>

//////////////
// Includes
#include "ClauseList.h"
#include "VariableSet.h"

#include <list>

/////////////
// Defines

////////////////////////
// Class Declarations
class ClConstraint;
class ClSimplexSolver;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Class Definitions
class VariableStruct {
 public:
  VariableStruct() : pFilterClause1(0), pFilterClause0(0), 
    pDeleteClause(0), myLen(0), pConstraint(0), eAssignment(NON_VALUE),
    bInSolver(false) {}
  ~VariableStruct();

  void vSetConstraint(ClConstraint * pcn);
  ClConstraint * pGetConstraint() { return pConstraint; }

  bool bIsConstrained() { return pConstraint != 0; }
  bool bIsInSolver() { return bInSolver; }

  int iConsSize();

  ClauseList xPositiveClauses; // stores list of clauses containing 
						 // positive occurence of this variable.
  ClauseList xNegativeClauses; // stores list of clauses containing 
						 // negative occurence of this  variable.
  Clause* pFilterClause1; // stores the clause filtering out "1" from 
					 // this domain
  Clause* pFilterClause0; // stores the clause filtering out "0" from 
					 // this domain
  Clause* pDeleteClause;  // I count on this indicating a branch point for
                          // learning.
  ClauseList xDeleteList;

  // Wolf - added to keep cheap track of constraint length
  int myLen;

  // Wolf - added this to allow variables tied to constraints. If this is
  //        null, no constraint. Otherwise, this variable (when set to true
  //        initiates a constraint.
  ClConstraint * pConstraint;

  // Wolf - Here's the solver that decides whether a given 
  // constraint set works:
  static ClSimplexSolver xSolver;
  static SATSolver * pSATSolver;

  // Wolf - each time an add fails, this stores the explanation
  //        of the failure; either a set of constraints which were
  //        active when the failure occurred and are mutually 
  //        inconsistent, or an empty set, meaning
  //        that no better explanation than "all active constraints"
  //        was available.
  static VariableSet xExplanation;

  DomainValue eGetAssignment() { return eAssignment; }

  void vFastUnsetAssignment(bool bConstrain) { 
    if (bConstrain) {
	 vUnsetAssignment();
	 return;
    }

    eAssignment = NON_VALUE;
  }

  void vUnsetAssignment() { 
    if (eAssignment == NON_VALUE)
	 return;

#ifndef NDEBUG
    bool bWasPos = eAssignment == 1;
#endif

    // Let non-constraining variables and false/unset vars off the hook:
    eAssignment = NON_VALUE;

    if (pConstraint != 0) {
	 if (bInSolver) {
	   _vRemoveConstraint();
	 }
	 else if (_xCache.back() == this) {
	   _xCache.pop_back();
	   _iCacheSize--;
	 }
	 else {
	   assert(!bWasPos || _iCacheSize == 0);
	 }
    }
  }

  // Returns 1 for contradiction, 0 for normal setting
  boolean bFastSetAssignment(DomainValue eAssignment_, bool bConstrain) { 
    if (bConstrain)
	 return bSetAssignment(eAssignment_);

    assert(eAssignment_ == 0 || eAssignment_ == 1);

    eAssignment = eAssignment_;

    return 0;
  }

  // Returns 1 for contradiction, 0 for normal setting
  boolean bSetAssignment(DomainValue eAssignment_) { 
    assert(eAssignment_ == 0 || eAssignment_ == 1);
#ifndef NDEBUG
    bool bTransitionOut = (eAssignment_ == 0 && eAssignment == 1);
#endif

    eAssignment = eAssignment_;

    if (pConstraint != 0) {
	 if (eAssignment == 0) {
	   if (bInSolver) {
		_vRemoveConstraint();
	   }
	   else if (_xCache.back() == this) {
		_xCache.pop_back();
		_iCacheSize--;
	   }
	   else {
		assert(!bTransitionOut || _xCache.empty());
	   }

	   return 0;
	 }
	 else if (bInSolver == 0 && eAssignment == 1) { // Need to add constraint
	   if (_bAutoFlush) {
		return _bAddConstraint();
	   }
	   else {
		_xCache.push_back(this);
		_iCacheSize++;
		return 0;
	   }
	 }
    }

    return 0;
  }

  static bool bFlush();  // Flush the cache of accumulated constraints. 
                         // returns 1 if a failure occurs in flushing.

  static bool bSetAutoFlush(bool bSetting) {
    if (bSetting) {
	 _bAutoFlush = true;
	 return bFlush();
    }
    else {
	 _bAutoFlush = false;
	 return 0;
    }
  }

  static bool bAutoFlushing() { return _bAutoFlush; }
  static int iCacheSize() { return _iCacheSize; }

 protected:

  // Wolf - since the constraint solver seems to be the dominating
  // time factor, I'm using a caching mechanism to add the
  // constraints. Note that with caching in place, an add will
  // _always_ return true. Only a flush will report failure.
  static list<VariableStruct *> _xCache;
  static int _iCacheSize;
  static bool _bAutoFlush;

  DomainValue eAssignment;
  boolean bInSolver;

  void _vRemoveConstraint();
  
  // returns 1 if constraint conflict occurs
  boolean _bAddConstraint();
};

class SATSolver {
public:
  // Tweak parameters:
  static
  double scoreCutoffFactor;          // how close to the best score
                                     // does a var's score need to be
                                     // to make it into the running
                                     // as the selected var [0-1]
  static
  double bestScoreFudgeFactor;       // within what factor of the best
                                     // score are all scores 
                                     // indistinguishable [0-1]
  static
  double constraintPreferenceFactor; // what factor should 
                                     // be added to constrained vars' 
                                     // scores

  static
  double constraintSizeFactor;       // score += size * size-factor
                                     // how important is the size of 
                                     // a constraint in determining its score.
  static
  bool fastUnitPropConstrained;      // should constrained vars be included
                                     // in fast unit propagation?
  static
  bool bExplanation;                 // use explanations in the LP solver?

  static
  bool bNaiveSolver;                 // should CBJ/learning be used?

  static
  bool bConstrainedVarsNegFirst;     // should constrained vars always be
                                     // set to 0 (false) first?

  static
  bool bUsePureLitElim;              // should pure lit elimination be used

  static
  int iConsMaxLearnSize;             // maximum sized clause to learn based
                                     // on a constraint conflict. 0 == none

  static
  int iLogicMaxLearnSize;            // maximum sized clause to learn based
                                     // on a logic conflict. 0 == none

  static
  bool bLearn;                       // Perform learning at all? (Does _not_
                                     // switch to naive if false!

  static
  bool bRelevance;                   // Perform relevance bounded learning?

  static
  bool bTimeLimit;                   // Respect the time limit?

  static
  int iMaxTime;                      // Time limit in hundredths of seconds

  static
  int iCachePolicy;                  // cache policy (settings in ipp.h, sadly)

  static
  int iCacheDepth;                   // associated depth (max acceptable, one
                                     // less than trigger depth)

public:
  SATSolver();

  void vReset();

  boolean bFindSolution(boolean& rFailed_);
  unsigned long lElapsedTime() { return _lElapsedTime;}

  boolean bGetValue(VariableID eVar);
  double dGetValue(int iVar);

  int iGetDepth() { return _iCurrentVariable; }
  int iBranchDepth() { return _iBranchDepth; }

  void vOutput(ostream & out);

private:
  // Initialization routines.
  boolean _bInitializeVariableInformation();
  inline void _vInitializeLargeClause(Clause*);
  inline void _vInitializeBinaryClause(Clause*);
  boolean _bInitializeUnaryClause(Clause*);

  void _vBackOut();

  VariableID _eTryAndPickBest(VariableSet& rTryVars_, boolean&);
  void _vFindBestSet(VariableSet&);
  void _vFillCandidateSet(double&, int&);
  double _iComputeScore();
  inline double _iCombineScores(double i1_, double i2_);

  boolean _bLoop(boolean&);
  boolean _bFunLoop(boolean&);

  boolean _bFastUnitPropagate();
  void _vFastBackup(VariableID);

  boolean _bUnitPropagate();

  boolean _bBackup();
  boolean _bNaiveBackup();
  boolean _bLearnBackup();

  inline boolean _bFilterWithClauseList(ClauseList&);
  void _vUnfilterWithClauseList(ClauseList&);
  void _vSatisfyWithClauseList(ClauseList&);

  void _vCreateBackupClauseFromContradictions();
  void _vCreateBackupClauseFromConstraintContradictions();
  void _vCreateBackupClauseFromLogicalContradictions();
  inline void _vCreateNewBackupClause();
  inline void _vLearn();

  inline boolean _bLabelVariable(VariableID, DomainValue);
  inline boolean _bFastLabelVariable(VariableID, DomainValue, bool bConstrain);
  inline void _vDeleteClauses(ClauseList& rClauseList_);

  // Private data
  struct tms buffer;

  unsigned long _lElapsedTime;

  unsigned long _lVariablesLabelled;

  VariablePosition _iCurrentVariable;
  int _iBranchDepth;
  VariableID _eCurrentID;
  int _iVariableCount;
  VariableID _eContradictionID;
  boolean _bLogicContradiction;
  boolean _bConsContradiction;

  VariableStruct _aVariableStruct[MAX_VARIABLES];
  VariableID _aPositionToID[MAX_VARIABLES]; // Maps the variables position to the variable ID
  double _aScore[MAX_VARIABLES];
  VariableID _aCandidates[MAX_VARIABLES];

  VariableSet _xUnitVariables0;  // Variables with unit sized domains containing 0
  VariableSet _xUnitVariables1;  // Variables with unit sized domains containing 1
  VariableSet _xUnlabeledSet;
  VariableSet _xBackupSet;
  VariableSet _xNegatedBackupSet;
  VariableSet _xPermaVariables;
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Inlines

inline double SATSolver::_iCombineScores(double i1_, double i2_)
{
  return (i1_ * i2_) + i1_ + i2_;
}

#endif // SATSolver_h

